import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.ArrayList;

/**
 * Liest ein Labyrinth aus einer Datei.
 * 
 * @author Bernd Gramlich
 * @version 2013-02-05
 */
public class Datenleser
{
    private static Datenleser datenleserSingleton;

    /*
     * Die Codezeichen, die wir im Labyrinth erwarten.
     */
    private final char WAND      = '#';
    private final char LEERRAUM  = '.';
    private final char EINGANG   = '@';
    private final char SCHATZ    = '$';
    private final char MARKENORD = '^';
    private final char MARKEWEST = '<';
    private final char MARKESUED = 'v';
    private final char MARKEOST  = '>';

    /**
     * Konstruktor.
     * 
     * Der Konstruktor ist private, weil er von aussen
     * nicht aufgerufen werden soll. Stattdessen soll
     * mit gibDatenleser () auf den Datenleser
     * zugegriffen werden.
     */
    private Datenleser ()
    {
    }

    /**
     * Gib den einzigen verfuegbaren Datenleser zurueck
     * (also den "datenleserSingleton"). Wenn der
     * Datenleser noch nicht erzeugt ist, erzeuge ihn.
     */
    public static Datenleser gibDatenleser ()
    {
        if (null == datenleserSingleton)
        {
            datenleserSingleton = new Datenleser ();
        }

        return datenleserSingleton;
    }

    /**
     * Lies die vom Benutzer gewuenschte Labyrinthdatei ein
     * und gib sie als ein zum Rechteck ergaenztes Feld von
     * Textzeichen aus.
     * 
     * Codierung:
     * WAND ist ein Stueck undurchdringliche Wand.
     * LEERRAUM ist ein Leerraum.
     * EINGANG ist der Ein- und Ausgang.
     * SCHATZ ist ein Schatz.
     * MARKENORD ist eine Marke, die nach Norden zeigt.
     * MARKEWEST  ist eine Marke, die nach Westen zeigt.
     * MARKESUED ist eine Marke, die nach Sueden zeigt.
     * MARKEOST ist eine Marke, die nach Osten zeigt.
     * 
     * Der Rand des Rechtecks wird als Wand interpretiert.
     * Das Rechteck ist so breit wie seine laengste Zeile. Zu
     * kurze Zeilen werden mit Leerraeumen ergaenzt. Sollte
     * mehr als ein EINGANG-Zeichen enthalten sein, so werden
     * alle EINGANG-Zeichen bis auf das erste durch Leerraeume
     * ersetzt. Sollte kein EINGANG-Zeichen vorhanden sein, so
     * wird der erste Leerraum als solches verwendet. Sollte
     * kein Leerraum vorhanden sein, so wird die erste Wand
     * dafuer verwendet. Sollte auch keine Wand vorhanden sein,
     * so wird der erste Schatz zum Eingang. Im aeussersten
     * Fall, dass die Datei kein einziges Zeichen enthaelt,
     * lautet die Ausgabe {{EINGANG}}.
     */
    public char[][] lies (File labyrinthDatei)
    {
        try
        {
            /*
             * Oeffne einen Leser fuer die Datei.
             */
            BufferedReader leser = new BufferedReader (new FileReader (labyrinthDatei));

            /*
             * Speichere die Eingabezeilen zunaechst in einer Liste.
             */
            ArrayList<String> eingabezeilen = new ArrayList<String> ();

            /*
             * Merke Dir die hoechste vorkommende Breite.
             */
            int breite = 0;

            /*
             * Lies alle Zeilen der Datei Stueck fuer Stueck ein.
             */
            while (leser.ready ())
            {
                String zeile = leser.readLine ();

                if (zeile.length () > breite)
                {
                    breite = zeile.length ();
                }

                eingabezeilen.add (zeile);
            }

            leser.close ();

            /*
             * Verlaengere alle zu kurzen Zeilen auf die
             * maximale Breite.
             */
            String leerkette = "";

            if (0 == breite)
            {
                breite = 1;
            }

            for (int j = 0; j < breite; j++)
            {
                leerkette = leerkette + ".";
            }

            for (int i = 0; i < eingabezeilen.size (); i++)
            {
                eingabezeilen.set (i, (eingabezeilen.get (i) + leerkette).substring (0, breite));
            }

            /*
             * Ermittle die Hoehe des Labyrinths.
             */
            int hoehe = eingabezeilen.size ();

            /*
             * Erstelle das zweidimensionale Zeichenkettenfeld,
             * das zurueckgegeben werden soll. Wenn es keine
             * Eingabezeilen gibt, erstelle ein Feld, das
             * einfach nur einen Eingang enthaelt.
             */
            char ausgabe[][] = {{EINGANG}};

            if (hoehe > 0)
            {
                ausgabe = new char[hoehe][breite];
            }

            int i = 0;
            int j;

            for (String zeile: eingabezeilen)
            {
                ausgabe[i++] = zeile.toCharArray ();
            }

            /*
             * Zaehle die einzelnen Codezeichen im Labyrinth.
             */
            int eingaenge  = 0;
            int waende     = 0;
            int leerraeume = 0;
            int schaetze   = 0;
            int markennord = 0;
            int markenwest = 0;
            int markensued = 0;
            int markenost  = 0;

            for (i = 0; i < hoehe; i++)
            {
                for (j = 0; j < breite; j++)
                {
                    switch (ausgabe[i][j])
                    {
                        case EINGANG:
                        if (0 == eingaenge)
                        {
                            eingaenge++;
                        }
                        else
                        {
                            ausgabe[i][j] = LEERRAUM;
                            leerraeume++;
                        }
                        break;

                        case LEERRAUM:
                        leerraeume++;
                        break;

                        case WAND:
                        waende++;
                        break;

                        case SCHATZ:
                        schaetze++;
                        break;
                        
                        /*
                         * Die gespeicherten Karten sind - wie frueher
                         * ueblich - nach Osten orientiert. Darum muessen
                         * wir die Marken entsprechen uebersetzen.
                         */
                        
                        case MARKENORD:
                        ausgabe[i][j] = MARKEWEST;
                        markenwest++;
                        break;
                        
                        case MARKEWEST:
                        ausgabe[i][j] = MARKESUED;
                        markensued++;
                        break;
                        
                        case MARKESUED:
                        ausgabe[i][j] = MARKEOST;
                        markenost++;
                        break;
                        
                        case MARKEOST:
                        ausgabe[i][j] = MARKENORD;
                        markennord++;
                        break;

                        default:
                        ausgabe[i][j] = LEERRAUM;
                        leerraeume++;
                        break;
                    }
                }
            }

            /*
             * Sorge dafuer, dass es einen Eingang gibt.
             */
            if (0 == eingaenge)
            {
                char suchzeichen;

                if (0 != leerraeume)
                {
                    suchzeichen = LEERRAUM;
                }
                else if (0 != markennord)
                {
                    suchzeichen = MARKENORD;
                }
                else if (0 != markenwest)
                {
                    suchzeichen = MARKEWEST;
                }
                else if (0 != markensued)
                {
                    suchzeichen = MARKESUED;
                }
                else if (0 != markenost)
                {
                    suchzeichen = MARKEOST;
                }
                else if (0 != waende)
                {
                    suchzeichen = WAND;
                }
                else
                {
                    suchzeichen = SCHATZ;
                }

                boolean erledigt = false;

                for (i = 0; i < hoehe; i++)
                {
                    for (j = 0; j < breite; j++)
                    {
                        if (!erledigt && ausgabe[i][j] == suchzeichen)
                        {
                            ausgabe[i][j] = EINGANG;
                            erledigt = true;
                        }
                    }
                }
            }

            return ausgabe;
        }
        catch (Exception fehler)
        {
            return null;
        }
    }
}
