import java.io.File;
import java.util.ArrayList;

/**
 * Ein Labyrinth mit dicken Waenden.
 * Das Labyrinth wird als zweidimensionales, rechteckiges
 * Feld von Textzeichen gespeichert.
 * 
 * @author Bernd Gramlich
 * @version 2013-02-05
 */
public class DickwandigesLabyrinth implements LabyrinthModel
{
    /*
     * Die Codezeichen, die wir im Labyrinth erwarten.
     */
    private final char WAND      = '#';
    private final char LEERRAUM  = '.';
    private final char EINGANG   = '@';
    private final char SCHATZ    = '$';
    private final char MARKENORD = '^';
    private final char MARKEWEST = '<';
    private final char MARKESUED = 'v';
    private final char MARKEOST  = '>';

    /*
     * Das Minimallabyrinth, das nur aus einem Eingang besteht.
     */
    private final char[][] minimalLabyrinth = {{EINGANG}};

    /*
     * Das dickwandige Labyrinth.
     */
    private char[][] labyrinth;

    /*
     * Die Breite und Hoehe des dickwandigen
     * Labyrinths.
     */
    private int breite;
    private int hoehe;

    /*
     * Die duennwandige, orientierte Ansicht des
     * Labyrinths, die wir an die Beobachter versenden.
     */
    private char[][] ansicht;

    /*
     * Die Seitenlaenge des quadratischen
     * Labyrinthausschnitts.
     */
    private int seitenlaenge;

    /*
     * Unsere Position im dickwandigen Labyrinth.
     */
    private int positionX;
    private int positionY;

    /*
     * Unsere Blickrichtung.
     */
    private Himmelsrichtung blickrichtung;

    /*
     * Die Anzahl der unentdeckten Schaetze.
     */
    private int schaetze;

    /*
     * Die Listen unserer Beobachter.
     */
    private ArrayList<PositionsBeobachter> positionsBeobachter;
    private ArrayList<SchatzBeobachter> schatzBeobachter;

    /**
     * Konstruktor.
     * Erzeuge ein Minimallabyrinth und siehe dafuer eine
     * Ansicht mit der genannten Seitenlaenge vor.
     */
    public DickwandigesLabyrinth (int dieseSeitenlaenge)
    {
        breite = 1;
        hoehe  = 1;

        labyrinth = minimalLabyrinth;

        positionX = 0;
        positionY = 0;
        blickrichtung = Himmelsrichtung.NORD;

        schaetze = 0;

        seitenlaenge = dieseSeitenlaenge;

        ansicht = new char [4*seitenlaenge+1][4*seitenlaenge+1];

        positionsBeobachter = new ArrayList<PositionsBeobachter> ();
        schatzBeobachter    = new ArrayList<SchatzBeobachter> ();
    }

    /**
     * Gehe im Labyrinth einen Schritt geradeaus.
     */
    public void gehe ()
    {
        /*
         * Bewege Dich auf das naechste Feld geradeaus.
         */
        switch (blickrichtung)
        {
            case NORD: positionY++; break;
            case WEST: positionX--; break;
            case SUED: positionY--; break;
            case OST:  positionX++; break;
        }
        
        /*
         * Wenn das Feld ein Schatzfeld ist, nimm den
         * Schatz (mache das Feld also zum Leerraum).
         */
        if (SCHATZ == gibLabyrinthinhalt (positionX, positionY))
        {
            labyrinth[positionX][positionY] = LEERRAUM;
            schaetze --;
            benachrichtigeSchatzBeobachter ();
        }
        
        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();
        
        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }

    /**
     * Pruefe, ob ein Schritt geradeaus moeglich ist.
     * 
     * return True genau dann, wenn ein Schritt geradeaus moeglich ist.
     */
    public boolean istFrei ()
    {
        /*
         * Ich ermittle die Labyrinthkachel vor mir.
         */
        char kachelVorMir = WAND;
        
        switch (blickrichtung)
        {
            case NORD: kachelVorMir = gibLabyrinthinhalt (positionX    , positionY + 1); break;
            case WEST: kachelVorMir = gibLabyrinthinhalt (positionX - 1, positionY    ); break;
            case SUED: kachelVorMir = gibLabyrinthinhalt (positionX    , positionY - 1); break;
            case OST:  kachelVorMir = gibLabyrinthinhalt (positionX + 1, positionY    ); break;
        }
        
        /*
         * Melde, ob die Labyrinthkachel vor mit keine Wand ist.
         */
        return (kachelVorMir != WAND);
    }

    /**
     * Drehe Dich im Labyrinth nach links.
     */
    public void dreheNachLinks ()
    {
        /*
         * Ich schwenke meinen Blick nach links.
         */
        switch (blickrichtung)
        {
            case NORD: blickrichtung = Himmelsrichtung.WEST; break;
            case WEST: blickrichtung = Himmelsrichtung.SUED; break;
            case SUED: blickrichtung = Himmelsrichtung.OST; break;
            case OST:  blickrichtung = Himmelsrichtung.NORD; break;
        }
        
        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();
        
        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }

    /**
     * Drehe Dich im Labyrinth nach rechts.
     */
    public void dreheNachRechts ()
    {
        /*
         * Ich schwenke meinen Blick nach rechts.
         */
        switch (blickrichtung)
        {
            case NORD: blickrichtung = Himmelsrichtung.OST; break;
            case OST:  blickrichtung = Himmelsrichtung.SUED; break;
            case SUED: blickrichtung = Himmelsrichtung.WEST; break;
            case WEST: blickrichtung = Himmelsrichtung.NORD; break;
        }
        
        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();
        
        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }
    
    /**
     * Setze an der augenblicklichen Position eine Marke.
     */
    public void setzeMarke ()
    {
        if (EINGANG != labyrinth[positionX][positionY])
        {
            /*
             * Falls ich nicht am Eingang bin, setze ich eine Marke, die
             * entgegen der Blickrichtung zeigt (also den Rueckweg weist).
             */
            switch (blickrichtung)
            {
                case NORD: labyrinth[positionX][positionY] = MARKESUED; break;
                case WEST: labyrinth[positionX][positionY] = MARKEOST;  break;
                case SUED: labyrinth[positionX][positionY] = MARKENORD; break;
                case OST:  labyrinth[positionX][positionY] = MARKEWEST; break;
            }
        
            /*
             * Erneuere die Labyrinthansicht.
             */
            erstelleAnsicht ();
        }
    }
    
    /**
     * Loesche die Marke an der augenblicklichen Position (sofern
     * dort eine ist).
     */
    public void loescheMarke ()
    {
        if (LEERRAUM != labyrinth[positionX][positionY])
        {
            /*
             * Wenn an der augenblicklichen Stelle kein Leerraum
             * ist, mache sie zum Leerraum.
             */
            labyrinth[positionX][positionY] = LEERRAUM;
            
            /*
             * Erneuere die Labyrinthansicht.
             */
            erstelleAnsicht ();
        }
    }

    /**
     * Pruefe, ob an der augenblicklichen Position eine Marke ist.
     * 
     * return True genau dann, wenn eine Marke vorhanden ist.
     */
    public boolean istMarke ()
    {
        boolean ausgabe = false;
        
        /*
         * Nur falls eine Marke vorhanden ist, die entgegen der Blickrichtung
         * zeigt, melden wir Erfolg.
         */
        switch (blickrichtung)
        {
            case NORD: ausgabe = (MARKESUED == labyrinth[positionX][positionY]); break;
            case WEST: ausgabe = (MARKEOST  == labyrinth[positionX][positionY]); break;
            case SUED: ausgabe = (MARKENORD == labyrinth[positionX][positionY]); break;
            case OST:  ausgabe = (MARKEWEST == labyrinth[positionX][positionY]); break;
        }
        
        return ausgabe;
    }
        
    /**
     * Lade das Labyrinth aus der Labyrinthdatei.
     * 
     * @param labyrinthDatei Die Datei, aus der das Labyrinth geladen werden soll.
     */
    public void ladeLabyrinth (File labyrinthDatei)
    {
        /*
         * Das Laden uebernimmt die Klasse Datenleser.
         */
        labyrinth = Datenleser.gibDatenleser ().lies (labyrinthDatei);

        /*
         * Wenn das Laden fehlschlaegt, geben wir keine
         * Fehlermeldung aus, sondern leben in einem
         * Minimallabyrinth.
         */
        if (null == labyrinth)
        {
            labyrinth = minimalLabyrinth;
        }

        /*
         * Wir verlassen uns darauf, dass das geladene
         * Labyrinth rechteckig ist (oder vom Datenleser
         * rechteckig gemacht wurde).
         */
        hoehe  = labyrinth.length;
        breite = labyrinth[0].length;
        
        /*
         * Ermittle den Eingang und die Anzahl der Schaetze.
         */
        positionX = -1;
        positionY = -1;
        
        schaetze = 0;
        
        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                switch (labyrinth[x][y])
                {
                    case EINGANG:
                    positionX = x;
                    positionY = y;
                    break;
                    
                    case SCHATZ:
                    schaetze++;
                    break;
                }
            }
        }
        
        /*
         * Pruefe nacheinander im Norden, Westen und Sueden,
         * ob dort keine Wand ist. Wenn ja, blicke dorthin.
         * Wenn nein, blicke nach Osten.
         */
        if (WAND != gibLabyrinthinhalt (positionX, positionY + 1))
        {
            blickrichtung = Himmelsrichtung.NORD;
        }
        else if (WAND != gibLabyrinthinhalt (positionX - 1, positionY))
        {
            blickrichtung = Himmelsrichtung.WEST;
        }
        else if (WAND != gibLabyrinthinhalt (positionX, positionY - 1))
        {
            blickrichtung = Himmelsrichtung.SUED;
        }
        else
        {
            blickrichtung = Himmelsrichtung.OST;
        }
        
        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();
        
        /*
         * Gib an die Beobachter weiter, wie viele Schaetze
         * im Labyrinth sind.
         */
        benachrichtigeSchatzBeobachter ();
    }
         
    /**
     * Nenne die Seitenlaenge der Labyrinthansicht.
     */
    public int gibSeitenlaenge ()
    {
        return seitenlaenge;
    }
     
    /**
     * Gib die aktuelle Ansicht des Labyrinths.
     */
    public char[][] gibLabyrinthansicht ()
    {
        return ansicht;
    }

    /**
     * Gib die aktuelle Anzahl von Schaetzen.
     */
    public int gibSchatzzahl ()
    {
        return schaetze;
    }

    /**
     * Speichere einen Positionsbeobachter.
     * 
     * @param b Der Positionsbeobachter.
     */
    public void speichereAlsBeobachter (PositionsBeobachter b)
    {
        positionsBeobachter.add (b);
    }

    /**
     * Speichere einen Schatzbeobachter.
     * 
     * @param b Der Schatzbeobachter.
     */
    public void speichereAlsBeobachter (SchatzBeobachter b)
    {
        schatzBeobachter.add (b);
    }

    /**
     * Loesche einen Positionsbeobachter.
     * 
     * @param b Der Positionsbeobachter.
     */
    public void loescheAlsBeobachter (PositionsBeobachter b)
    {
        positionsBeobachter.remove (b);
    }

    /**
     * Loesche einen Schatzbeobachter.
     * 
     * @param b Der Schatzbeobachter.
     */
    public void loescheAlsBeobachter (SchatzBeobachter b)
    {
        schatzBeobachter.remove (b);
    }

    /**
     * Erstelle die duennwandige Ansicht.
     */
    private void erstelleAnsicht ()
    {
        /*
         * Fuelle die Ansicht zunaechst mit Leerraeumen.
         */
        for (int i = 0; i <= 4*seitenlaenge; i++)
        {
            for (int j = 0; j <= 4*seitenlaenge; j++)
            {
                ansicht[i][j] = LEERRAUM;;
            }
        }
        /*
         * Bestimme die Grenzen und die Schrittweiten der x-
         * und der y-Koordinaten.
         */
        int x1, x2, dx;
        int y1, y2, dy;

        switch (blickrichtung)
        {
            case NORD:
            x1 = positionX - seitenlaenge;
            x2 = positionX + seitenlaenge;
            dx = +1;
            y1 = positionY - seitenlaenge;
            y2 = positionY + seitenlaenge;
            dy = +1;
            break;

            case WEST:
            x1 = positionY - seitenlaenge;
            x2 = positionY + seitenlaenge;
            dx = +1;
            y1 = positionX + seitenlaenge;
            y2 = positionX - seitenlaenge;
            dy = -1;
            break;

            case SUED:
            x1 = positionX + seitenlaenge;
            x2 = positionX - seitenlaenge;
            dx = -1;
            y1 = positionY + seitenlaenge;
            y2 = positionY - seitenlaenge;
            dy = -1;
            break;

            default: // d. h. case OST:
            x1 = positionY + seitenlaenge;
            x2 = positionY - seitenlaenge;
            dx = -1;
            y1 = positionX - seitenlaenge;
            y2 = positionX + seitenlaenge;
            dy = +1;
            break;
        }
        
        /*
         * Verwende fuer die Ansicht die Koordinaten relativen
         * Koordinaten (u|w) zur Unterscheidung von dern
         * absoluten Koordinaten (x|y).
         */
        int u, v;

        /*
         * Trage alle Waende, Eingaenge und Schaetze korrekt in
         * die Ansicht ein.
         */
        u = -seitenlaenge;
        
        for (int x = x1; x != x2 + dx; x += dx)
        {
            v = -seitenlaenge;
            
            for (int y = y1; y != y2 + dy; y += dy)
            {
                char inhalt = WAND;
                
                switch (blickrichtung)
                {
                    case NORD: case SUED: inhalt = gibLabyrinthinhalt (x, y); break;
                    case OST:  case WEST: inhalt = gibLabyrinthinhalt (y, x); break;
                }
                
                switch (inhalt)
                {
                    /*
                     * Gruppiere vier Waende um ein Wandelement des
                     * dickwandigen Labyrinths herum.
                     */
                    case WAND:
                    setzeWand (index (u) + 1, index (v)    );
                    setzeWand (index (u)    , index (v) + 1);
                    setzeWand (index (u) - 1, index (v)    );
                    setzeWand (index (u)    , index (v) - 1);
                    break;
                    
                    /*
                     * Uebertrage den Eingang.
                     */
                    case EINGANG:
                    ansicht[index (u)    ][index (v)    ] = EINGANG;
                    break;
                    
                    /*
                     * Uebertrage die Schaetze.
                     */
                    case SCHATZ:
                    ansicht[index (u)    ][index (v)    ] = SCHATZ;
                    break;
                    
                    /*
                     * Uebertrage die Marken in ihrer richtigen Orientierung.
                     */
                    case MARKENORD:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case SUED: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case OST : ansicht[index (u)][index (v)] = MARKEWEST; break;
                    }
                    break;
                    
                    case MARKEWEST:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKEWEST; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case SUED: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case OST : ansicht[index (u)][index (v)] = MARKESUED; break;
                    }
                    break;
                    
                    case MARKESUED:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKEWEST;  break;
                        case SUED: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case OST : ansicht[index (u)][index (v)] = MARKEOST;  break;
                    }
                    break;
                    
                    case MARKEOST:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case WEST: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case SUED: ansicht[index (u)][index (v)] = MARKEWEST; break;
                        case OST : ansicht[index (u)][index (v)] = MARKENORD; break;
                    }
                    break;    
                }
                v++;
            }
            u++;
        }
        
        benachrichtigePositionsBeobachter ();
    }
    
    /**
     * Pruefe, ob das Spielende erreicht ist.
     */
    private void pruefeSpielende ()
    {
        /*
         * Wenn ich alle Schaetze eingesammelt habe und zurueck
         * auf dem Eingang bin, dann vermindere ich die
         * Schatzzahl nochmals um 1, um anzuzeigen, dass das
         * Spiel zu Ende ist.
         */
        if (EINGANG == gibLabyrinthinhalt (positionX, positionY) && 0 == schaetze)
        {
            schaetze --;
            benachrichtigeSchatzBeobachter ();
        } 
    }

    /**
     * Nenne den Inhalt des Labyrinths an der Position
     * (x|y). Wenn die Position unmittelbar jenseits des
     * Labyrinthrandes ist, tun wir so, als waere dort
     * eine Wand. Wenn die Position noch weiter ausserhalb
     * des Labyrints liegt, tun wir so, als waere dort ein
     * Leerraum.
     */
    private char gibLabyrinthinhalt (int x, int y)
    {
        if (0 <= x && x < hoehe)
        {
            if (0 <= y && y < breite)
            {
                return labyrinth[x][y];
            }
            else if (-1 == y || breite == y)
            {
                return WAND;
            }
            else
            {
                return LEERRAUM;
            }
        }
        else if (-1 == x || hoehe == x)
        {
            if (0 <= y && y < breite)
            {
                return WAND;
            }
            else
            {
                return LEERRAUM;
            }
        }
        else
        {
            return LEERRAUM;
        }
    }

    /**
     * Rechne eine Koordinate des Labyrinths in einen
     * Feldindex der Ansicht um.
     */
    private int index (int k)
    {
        return 2*seitenlaenge - 2*k;
    }
    
    /**
     * Setze an der angegebenen Stelle eine Wand in die Ansicht.
     */
    private void setzeWand (int i, int j)
    {
        /*
         * Pruefe, ob die Stelle innerhalb der Ansicht liegt.
         */
        if (0 <= Math.min (i, j) && Math.max (i, j) <= 4*seitenlaenge)
        {
            if (WAND != ansicht[i][j])
            {
                /*
                 * Wenn an der Stelle noch keine Wand ist, setze eine.
                 */
                ansicht[i][j] = WAND;
            }
            else
            {
                /*
                 * Wenn an der Stelle bereits eine Wand ist, loesche
                 * sie (denn dann stossen dort zwei Wandelemente
                 * aneinander).
                 */
                ansicht[i][j] = LEERRAUM;
            }
        }   
    }

    /**
     * Benachrichtige Positionsbeobachter, dass sich an der Position
     * im Labyrinth etwas geaendert hat.
     */
    private void benachrichtigePositionsBeobachter ()
    {
        for (PositionsBeobachter b: positionsBeobachter)
        {
            b.aktualisiereLabyrinthansicht ();
        }
    }

    /**
     * Benachrichtige Schatzbeobachter, dass sich an der Anzahl der
     * Schaetze im Labyrinth etwas geaendert hat.
     */
    private void benachrichtigeSchatzBeobachter ()
    {
        for (SchatzBeobachter b: schatzBeobachter)
        {
            b.aktualisiereSchatzzahl ();
        }
    }

    private enum Himmelsrichtung
    {
        NORD,
        WEST,
        SUED,
        OST,
        }
    }
